@@ -8,11 +8,12 @@ from urllib.parse import urlparse
 
 import boto3
 import botocore
-
+import logging
 bucket = 'sagemaker-us-west-2-079310353266'
 prefix = 'sagemaker/sagemaker-xgboost-lending/'
 s3_client = boto3.client('s3')
 
+LOGGER = logging.getLogger(__name__)
 
 class S3Uri:
     """bucket, object_key for an S3 uri."""
@@ -39,11 +40,13 @@ class S3Uri:
             # using default credentials
             # if not auth_context, assume that the credentials are set up
             # elsewhere viz., envvars, credentials file etc.
+            LOGGER.info("Gettting credentials from the boto3 session")
             self.s3 = boto3.resource('s3', endpoint_url=endpoint_url)
         else:
             # there are no credentials.. Maybe, the user is expecting to
             # download from a public S3 bucket?
             # using no credentials
+            LOGGER.info("No credentials provided using public bucket")
             cnfg = botocore.config.Config(signature_version=botocore.UNSIGNED)
             self.s3 = boto3.resource('s3', endpoint_url=endpoint_url, config=cnfg)
 
