@@ -7,6 +7,7 @@ import json
 import logging
 import re
 import textwrap
+from dataclasses import asdict, dataclass
 from pathlib import Path
 from typing import (
     Any,
@@ -24,129 +25,19 @@ from typing import (
 import numpy as np
 import pandas as pd
 import pandas.api.types
+import pyarrow as pa
 
 from ._version import __version__
+from .utils.exceptions import MalformedSchemaException
+from .utils.formatting import prettyprint_number, validate_sanitized_names
+from .utils.general_checks import (
+    is_greater_than_max_value,
+    is_less_than_min_value,
+    type_enforce,
+)
+from .utils.pandas import is_datetime
 
-# from .utils import _type_enforce, MalformedSchemaException
-
-MAX_NAME_LEN = 60
 LOG = logging.getLogger()
-TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M:%S.%f'
-
-
-# should be in utils:
-def _type_enforce(param_name: str, param_val, required_type):
-    """
-    :param param_name: Name of parameter being type enforced. Used for valid error
-    :param param_val: Value being type enforced
-    :param required_type: Casted type that is being enforced
-    """
-    if isinstance(param_val, required_type):
-        return param_val  # should perhaps be required_type(param_val)?
-    else:
-        try:
-            return required_type(param_val)
-        except ValueError:
-            raise ValueError(
-                f'Parameter `{param_name}` must be of type {required_type}'
-            )
-
-
-def is_datetime(pandas_series):
-    """
-    Return true if Series contains all valid dates.
-    """
-    # if pandas_series contains valid timestamp series to_datetime will not
-    # throw an exception, which means all rows has timestamp.
-    try:
-        pd.to_datetime(pandas_series, format=TIMESTAMP_FORMAT, errors='raise')
-        return True
-    except Exception:
-        return False
-
-
-# should be in utils:
-class MalformedSchemaException(Exception):
-    def __init__(self, message='Core Fiddler Object has Malformed Schema'):
-        self.message = message
-        super().__init__(self.message)
-
-
-def is_greater_than_max_value(value, limit, epsilon=1e-9):
-    """Check if 'value' is significantly larger than 'limit'.
-    Where significantly means theres a greater difference than epsilon.
-    """
-    max_arg = abs(max(value, limit))
-    tolerance = max_arg * epsilon
-    # make sure that a is "significantly" smaller than b before declaring it
-    # identified in DI for hired
-    return value - tolerance > limit
-
-
-def is_less_than_min_value(value, limit, epsilon=1e-9):
-    """Check if 'value' is significantly smaller than 'limit'.
-    Where significantly means theres a greater difference than epsilon.
-    """
-    max_arg = abs(max(value, limit))
-    tolerance = max_arg * epsilon
-    # make sure that a is "significantly" smaller than b before declaring it
-    # identified in DI for hired
-    return value + tolerance < limit
-
-
-def name_check(name: str, max_length: int):
-    if not name:
-        raise ValueError(f'Invalid name {name}')
-    if len(name) > max_length:
-        raise ValueError(f'Name longer than {max_length} characters: {name}')
-    if len(name) < 2:
-        raise ValueError(f'Name must be at least 2 characters long: {name}')
-    if not bool(re.match('^[a-z0-9_]+$', name)):
-        raise ValueError(
-            f'Name must only contain lowercase letters, '
-            f'numbers or underscore: {name}'
-        )
-    if name.isnumeric():
-        raise ValueError(
-            f'Name must contain at least one ' f'alphabetical character: {name}'
-        )
-
-
-def compute_hash(string):
-    hash_value = 0
-    if not string:
-        return hash_value
-    for c in string:
-        hash_value = ((hash_value << 5) - hash_value) + ord(c)
-        hash_value = hash_value & 0xFFFFFFFF  # Convert to 32bit integer
-    return str(hash_value)
-
-
-def sanitized_name(name):
-    if name.isnumeric():
-        name = f'_{name}'
-    name = re.sub(r'[^a-zA-Z0-9_]', '_', name).lower()
-    if len(name) > 63:
-        suffix = f'_{compute_hash(name)}'
-        name = f'{name[0: 63-len(suffix)]}{suffix}'
-    return name
-
-
-def validate_sanitized_names(columns, sanitized_name_dict):
-    if not columns:
-        return
-    if isinstance(columns, str):
-        columns = [columns]
-    for column in columns:
-        sname = sanitized_name(column.name)
-        if sname in sanitized_name_dict:
-            other_name = sanitized_name_dict.get(sname)
-            raise ValueError(
-                f'Name conflict, {column.name} and {other_name} '
-                f'both maps to {sname}'
-            )
-        else:
-            sanitized_name_dict[sname] = column.name
 
 
 class IntegrityViolationStatus(NamedTuple):
@@ -193,6 +84,13 @@ class MonitoringViolationType(enum.Enum):
     WARNING = 'warning'
 
 
+@enum.unique
+class HigherLevelAggregates(enum.Enum):
+    """Supported Higher Level Aggregates"""
+
+    TOP_K = 'top K'
+
+
 @enum.unique
 class FiddlerEventColumns(enum.Enum):
     OCCURRED_AT = '__occurred_at'
@@ -334,7 +232,8 @@ class BuiltInMetrics(enum.Enum):
     F1_SCORE = 'f1_score'
     AUC = 'auc'
 
-    def get_binary():
+    @classmethod
+    def get_binary(cls) -> List['BuiltInMetrics']:
         'return binary classification metrics'
         return [
             BuiltInMetrics.ACCURACY,
@@ -344,11 +243,13 @@ class BuiltInMetrics(enum.Enum):
             BuiltInMetrics.AUC,
         ]
 
-    def get_multiclass():
+    @classmethod
+    def get_multiclass(cls) -> List['BuiltInMetrics']:
         'return multiclass classification metrics'
         return [BuiltInMetrics.ACCURACY, BuiltInMetrics.LOG_LOSS]
 
-    def get_regression():
+    @classmethod
+    def get_regression(cls) -> List['BuiltInMetrics']:
         'return regression metrics'
         return [
             BuiltInMetrics.R2,
@@ -368,6 +269,15 @@ class ModelInputType(enum.Enum):
     MIXED = 'mixed'
 
 
+@enum.unique
+class DeploymentType(str, enum.Enum):
+    EXECUTOR = 'executor'
+    FAR = 'far'
+    NO_MODEL = 'no-model'
+    PREDICTOR = 'predictor'
+    SURROGATE = 'surrogate'
+
+
 class AttributionExplanation(NamedTuple):
     """The results of an attribution explanation run by the Fiddler engine."""
 
@@ -420,7 +330,7 @@ class MulticlassAttributionExplanation(NamedTuple):
         """Converts a deserialized JSON format into an
         MulticlassAttributionExplanation object"""
         return cls(
-            classes=tuple(deserialized_json.keys()),
+            classes=cast(Tuple[str], tuple(deserialized_json.keys())),
             explanations={
                 label_class: AttributionExplanation.from_dict(explanation_dict)
                 for label_class, explanation_dict in deserialized_json.items()
@@ -473,30 +383,20 @@ class ModelDeploymentParams:
         return res
 
 
+@dataclass
 class DeploymentOptions:
-    def __init__(
-        self,
-        deployment_type: Optional[
-            str
-        ] = 'surrogate',  # model type. one of {'surrogate', 'no-model', 'predictor', 'executor'}
-        image_uri: Optional[str] = None,  # image to be used for newly uploaded model
-        namespace: Optional[str] = 'default',  # kubernetes namespace
-        port: Optional[int] = 5100,  # port on which model is served
-        replicas: Optional[int] = 1,  # number of replicas
-        cpus: Optional[float] = 0.25,  # number of CPU cores
-        memory: Optional[str] = '128m',  # amount of memory required.
-        gpus: Optional[int] = 0,  # number of GPU cores
-        await_deployment: Optional[bool] = True,  # wait for deployment
-    ):
-        self.deployment_type = deployment_type
-        self.image_uri = image_uri
-        self.namespace = namespace
-        self.port = port
-        self.replicas = replicas
-        self.cpus = cpus
-        self.memory = memory
-        self.gpus = gpus
-        self.await_deployment = await_deployment
+    deployment_type: str = DeploymentType.SURROGATE
+    image: str = None  # image to be used for newly uploaded model
+    namespace: Optional[str] = None  # kubernetes namespace
+    port: int = 5100  # port on which model is served
+    replicas: int = 1  # number of replicas
+    cpus: float = 0.25  # number of CPU cores
+    memory: str = '128m'  # amount of memory required.
+    gpus: int = 0  # number of GPU cores
+    await_deployment: bool = True  # wait for deployment
+
+    def to_dict(self):
+        return asdict(self)
 
 
 class Column:
@@ -583,10 +483,31 @@ class Column:
     def get_pandas_dtype(self):
         """Converts the data_type field to a Pandas-friendly form."""
         self._raise_on_bad_categorical()
+
         if self.data_type.value == DataType.CATEGORY.value:
             return pandas.api.types.CategoricalDtype(self.possible_values)
+
+        return self.data_type.value
+
+    def get_arrow_field(self) -> pa.Field:
+        """Converts the data_type field to a pyarrow field"""
+
+        if self.data_type == DataType.CATEGORY:
+            field_type = pa.dictionary(index_type=pa.int64(), value_type=pa.string())
+        elif self.data_type == DataType.INTEGER:
+            field_type = pa.int64()
+        elif self.data_type == DataType.FLOAT:
+            field_type = pa.float64()
+        elif self.data_type == DataType.BOOLEAN:
+            field_type = pa.bool_()
+        elif self.data_type == DataType.STRING:
+            field_type = pa.string()
         else:
-            return self.data_type.value
+            raise ValueError(
+                f'Not able to map data type to arrow field - {self.data_type}'
+            )
+
+        return pa.field(self.name, field_type)
 
     def to_dict(self) -> Dict[str, Any]:
         """Converts this object to a more JSON-friendly form."""
@@ -643,17 +564,13 @@ class Column:
             possible_types = tuple(set(type(v) for v in self.possible_values))
             return not isinstance(value, possible_types)
 
-    def violation_of_nullable(self, value, fall_back=None):
-        if fall_back:
-            if self.name in fall_back:
-                value = None if value in fall_back[self.name] else value
-
+    def violation_of_nullable(self, value):
         if self.is_nullable is not None and self.is_nullable is False:
             return Column._value_is_na_or_none(value)
         return False
 
-    def check_violation(self, value, fall_back=None):
-        if self.violation_of_nullable(value, fall_back):
+    def check_violation(self, value):
+        if self.violation_of_nullable(value):
             return IntegrityViolationStatus(True, False, False)
         if self.violation_of_type(value):
             return IntegrityViolationStatus(False, True, False)
@@ -702,7 +619,7 @@ class DatasetInfo:
     ):
         self.display_name = display_name
         self.dataset_id = dataset_id
-        self.columns = columns
+        self.columns = DatasetInfo._datatype_check(columns)
         self.files = files if files is not None else list()
         self.misc = kwargs
 
@@ -715,6 +632,32 @@ class DatasetInfo:
         }
         return {**res, **self.misc}
 
+    def get_pandas_dtypes(self) -> Dict:
+        """
+        Convert dataset info columns data types to pandas compatible data types
+        :return: Dictionary of pandas data types for dataset columns
+        """
+
+        dtypes = {}
+
+        for column in self.columns:
+            dtypes[column.name] = column.get_pandas_dtype()
+
+        return dtypes
+
+    def get_arrow_schema(self) -> pa.Schema:
+        """
+        Convert dataset info columns data types to pyarrow compatible schema
+        :return: pyarrow compatible schema
+        """
+
+        fields = []
+
+        for column in self.columns:
+            fields.append(column.get_arrow_field())
+
+        return pa.schema(fields)
+
     def get_column_names(self) -> List[str]:
         """Returns a list of column names."""
         return [column.name for column in self.columns]
@@ -739,7 +682,28 @@ class DatasetInfo:
         )
 
     @staticmethod
-    def _datatype_from_pandas_dtype(pandas_col) -> DataType:
+    def _datatype_check(columns):
+        """
+        # loop through all columns
+        # if the column datatype is numpy datatype, transform to python type
+        """
+        for column in columns:
+            column.value_range_max = DatasetInfo._transform_type(column.value_range_max)
+            column.value_range_min = DatasetInfo._transform_type(column.value_range_min)
+        return columns
+
+    @staticmethod
+    def _transform_type(val):
+        if isinstance(val, np.bool_):
+            val = bool(val)
+        if isinstance(val, np.integer):
+            val = int(val)
+        if isinstance(val, np.floating):
+            val = float(val)
+        return val
+
+    @staticmethod
+    def datatype_from_pandas_dtype(pandas_col) -> DataType:
         dtype_obj = pandas_col.dtype
         # catch uint8 one-hot encoded variables
         if dtype_obj.name == 'uint8' and pandas_col.nunique() < 3:
@@ -866,7 +830,14 @@ class DatasetInfo:
 
     @staticmethod
     def _calculate_stats_for_col(column_name, column_series, max_inferred_cardinality):
-        column_dtype = DatasetInfo._datatype_from_pandas_dtype(column_series)
+        # @TODO Automatically drop the empty column with warning and proceed instead of aborting the upload
+        if column_series.isna().all():
+            raise ValueError(
+                f'Column {column_name} is empty. '
+                f'Please remove it and re-upload the dataset.'
+            )
+
+        column_dtype = DatasetInfo.datatype_from_pandas_dtype(column_series)
 
         # infer categorical if configured to do so
         if (
@@ -961,7 +932,7 @@ class DatasetInfo:
             if col_type.value == DataType.CATEGORY.value:
                 assert a_column.possible_values is not None  # nosec
                 assert b_column.possible_values is not None  # nosec
-                possible_values = list(
+                possible_values: Optional[List[Any]] = list(
                     set(a_column.possible_values) | set(b_column.possible_values)
                 )
             else:
@@ -1098,7 +1069,6 @@ class ModelInfo:
     :param group_by: [Optional] A string representing the column name performance metrics have
      to be group by with for performance metrics computation. This have to be given for Ranking for MAP
       and NDCG computations. For ranking models, it represents the query/session id column.
-    :param fall_back: A Dictionary with the column name as key and a list of fall back values as value.
     :param **kwargs: Additional information about the model to store as `misc`.
     """
 
@@ -1191,6 +1161,7 @@ class ModelInfo:
                     f'The preferred_explanation_method specified ({preferred_explanation_method}) could not be found in the built-in explanation methods ({BUILT_IN_EXPLANATION_NAMES})'
                 )
 
+
         # Prevent the user from specifying an invalid schema for fall back value
         if self.fall_back is not None:
             for k, v in self.fall_back.items():
@@ -1489,7 +1460,8 @@ class ModelInfo:
         :param mlflow_params: MLFlow parameters.
         :param outputs: model output names, if multiclass classification, must be a sequence in the same order as categorical_target_class_details.
             If binary classification, must be a single name signifying the probability of the positive class. # TODO: bulletproofing
-            If regression/ranking, must be a dictionary of the form {column_name: (min_value, max_value)}
+            If regression/ranking, must be a dictionary of the form {column_name: (min_value, max_value)}, or a
+            dictionary with empty range {column_name:[]} if column_name can be found in the dataset
         :param categorical_target_class_details: specify the output categories of your model (only applicable for classification and ranking models)
             This parameter *must* be provided for multiclass, binary-classification and ranking models where
             the target is of type CATEGORY. It is optional for binary classification with BOOLEAN targets, and ignored for regression.
@@ -1736,7 +1708,7 @@ class ModelInfo:
                 raise ValueError(
                     f'Invalid arguments: model_task is specified as {model_task.value} but inferred as '
                     f'{inferred_type.value} due to the type of Argument '
-                    f'{"outputs("+ str(type(outputs)) if outputs else "target("+ str(target_column.data_type)}). '
+                    f'{"outputs(" + str(type(outputs)) if outputs else "target(" + str(target_column.data_type)}). '
                     '[USAGE for Argument outputs]: If multiclass classification, must specify a sequence in the same order as '
                     'categorical_target_class_details: [cls_1, cls_2 ... cls_n]; if binary classification, must be a '
                     'single name signifying the probability of the positive class: pos_cls_name; if regression or '
@@ -1855,6 +1827,7 @@ class ModelInfo:
                     )
             else:
                 LOG.info('Using inferred positive class.')
+
         if model_task.value == ModelTask.RANKING.value:
             if target_column.data_type.value not in ['float', 'int']:
                 if categorical_target_class_details[0] is None:
@@ -1985,6 +1958,11 @@ class ModelInfo:
         framework_info = (
             f'  framework: {self.framework}\n' if self.framework is not None else ''
         )
+
+        fall_back_info = (
+            f'  fall_back: {self.fall_back}\n' if self.fall_back is not None else ''
+        )
+
         misc_info = json.dumps(self.misc, indent=2)
         target_info = (
             f"<hr>targets:{summary_dict['targets']._repr_html_()}"
@@ -2001,9 +1979,11 @@ class ModelInfo:
             if self.metadata is not None
             else ''
         )
+
         fall_back_info = (
             f'  fall_back: {self.fall_back}\n' if self.fall_back is not None else ''
         )
+
         return (
             f'<div style="border: thin solid rgb(41, 57, 141); padding: 10px;">'
             f'<h3 style="text-align: center; margin: auto;">ModelInfo\n</h3><pre>'
@@ -2016,13 +1996,13 @@ class ModelInfo:
             f'  custom_explanation_names: {self.custom_explanation_names}\n'
             f'{algorithm_info}'
             f'{framework_info}'
+            f'{fall_back_info}'
             f'  misc: {misc_info}</pre>'
             f'{target_info}'
             f"<hr>inputs:{summary_dict['inputs']._repr_html_()}"
             f"<hr>outputs:{summary_dict['outputs']._repr_html_()}"
             f'{decisions_info}'
             f'{metadata_info}'
-            f'{fall_back_info}'
             f'</div>'
         )
 
@@ -2038,20 +2018,20 @@ class ModelInfo:
 
         metadata_info = (
             f'  metadata:\n'
-            f"{textwrap.indent(repr(summary_dict['metadata']),  '    ')}"
+            f"{textwrap.indent(repr(summary_dict['metadata']), '    ')}"
             if self.metadata is not None
             else ''
         )
 
         decisions_info = (
             f'  decisions:\n'
-            f"{textwrap.indent(repr(summary_dict['decisions']),  '    ')}"
+            f"{textwrap.indent(repr(summary_dict['decisions']), '    ')}"
             if self.decisions is not None
             else ''
         )
 
         target_info = (
-            f'  targets:\n' f"{textwrap.indent(repr(summary_dict['targets']),  '    ')}"
+            f'  targets:\n' f"{textwrap.indent(repr(summary_dict['targets']), '    ')}"
             if self.targets is not None
             else ''
         )
@@ -2138,9 +2118,9 @@ class SegmentInfo:
         """
         schema-version 0.1
         """
-        _type_enforce('project_id', project_id, str)
-        _type_enforce('model_id', model_id, str)
-        _type_enforce('segment_id', segment_id, str)
+        type_enforce('project_id', project_id, str)
+        type_enforce('model_id', model_id, str)
+        type_enforce('segment_id', segment_id, str)
         self.schema_version = CURRENT_SCHEMA_VERSION
         self.creator_version = f'FIDDLER-CLIENT-v{__version__}'
         self.project = project_id
@@ -2159,6 +2139,9 @@ class SegmentInfo:
         if metrics is None:
             self.metrics = []
         else:
+            raise MalformedSchemaException(
+                f'Segment specific metrics have not yet been implemented (schema-version {CURRENT_SCHEMA_VERSION}).'
+            )
             self.metrics = metrics
 
     def get_filter(self):
@@ -2208,11 +2191,11 @@ class SegmentInfo:
             segment_id=deserialized_json['segment_id'],
             filter=deserialized_json['filter'],
             subsegments=deserialized_json['subsegments'],
-            metrics=deserialized_json['metrics'],
             description=deserialized_json['description'],
         )
         seg_info.schema_version = deserialized_json['schema_version']
         seg_info.creator_version = deserialized_json['creator_version']
+        seg_info.metrics = deserialized_json['metrics']
         return seg_info
         # TODO: bulletproofing to check additional keys don't exist
 
@@ -2227,10 +2210,6 @@ class SegmentInfo:
             )
         result = result and self._validate_filter_format(model_info)
         result = result and self._validate_subsegments_format(model_info)
-        if len(self.metrics) == 0:
-            raise MalformedSchemaException(
-                'Cannot create segment with no associated metrics.'
-            )
         result = result and self._validate_metrics_format(model_info)
         return result
 
@@ -2320,8 +2299,15 @@ class SegmentInfo:
         """
         Validate metrics according to schema-version 0.1.
         """
+        if len(self.metrics) != 0:
+            raise MalformedSchemaException(
+                f'SegmentInfo cannot yet specify metrics (schema-version {CURRENT_SCHEMA_VERSION}).'
+            )
+        else:
+            return True
         # TODO: allow generic metrics like traffic count, etc
         valid_metrics = []
+        # ToDo: add ranking
         if model_info.model_task == ModelTask.BINARY_CLASSIFICATION:
             valid_metrics += [m.value for m in BuiltInMetrics.get_binary()]
         elif model_info.model_task == ModelTask.MULTICLASS_CLASSIFICATION:
@@ -2363,7 +2349,8 @@ def _summary_dataframe_for_columns(
     column_dtypes = []
     n_possible_values = []
     is_nullable = []
-    mins, maxes = [], []
+    mins: List[Any] = []
+    maxes: List[Any] = []
     for column in columns:
         column_names.append(column.name)
         column_dtypes.append(column.data_type.name)
@@ -2380,12 +2367,12 @@ def _summary_dataframe_for_columns(
             maxes.append(None)
         else:
             min_str = (
-                _prettyprint_number(column.value_range_min)
+                prettyprint_number(column.value_range_min)
                 if column.value_range_min is not None
                 else '*'
             )
             max_str = (
-                _prettyprint_number(column.value_range_max)
+                prettyprint_number(column.value_range_max)
                 if column.value_range_max is not None
                 else '*'
             )
@@ -2405,8 +2392,3 @@ def _summary_dataframe_for_columns(
             'value_range': value_range,
         }
     )
-
-
-def _prettyprint_number(number, n_significant_figures=4):
-    n_digits = len(f'{number:.0f}')
-    return f'{round(number, n_significant_figures - n_digits):,}'
