@@ -15,7 +15,7 @@ class ModelInfoValidator:
     def validate_categoricals(self, modify=False):
         for col in self.model_info.inputs:
             possible_values_floats = None
-            if col.data_type == DataType.CATEGORY:
+            if col.data_type == DataType.CATEGORY.value:
                 try:
                     possible_values_floats = [
                         str(float(raw_val)) for raw_val in col.possible_values
@@ -24,13 +24,9 @@ class ModelInfoValidator:
                     pass
             if possible_values_floats is not None:
                 if not modify:
-                    for f_val, c_val in zip(
-                        possible_values_floats, col.possible_values
-                    ):
+                    for f_val, c_val in zip(possible_values_floats, col.possible_values):
                         if str(f_val) != str(c_val):
-                            raise ValueError(
-                                f'Categoricals have invalid values: {f_val} != {c_val}'
-                            )
+                            raise ValueError(f'Categoricals have invalid values: {f_val} != {c_val}')
                 else:
                     col.possible_values = possible_values_floats
         pass
@@ -51,6 +47,7 @@ class ModelInfoValidator:
         if len(self.model_info.outputs) < 1:
             raise ValueError('Model outputs not defined')
 
+
         if self.model_info.model_task == ModelTask.REGRESSION:
             self.validate_regression_model()
         elif self.model_info.model_task == ModelTask.BINARY_CLASSIFICATION:
