@@ -1,5 +1,7 @@
 from datetime import datetime
+from typing import List
 
+from . import constants
 from .core_objects import (
     DataType,
     EventTypes,
@@ -9,14 +11,17 @@ from .core_objects import (
     MonitoringViolation,
     MonitoringViolationType,
 )
-from .utils import TIMESTAMP_FORMAT, formatted_utcnow, is_int_type, pad_timestamp
+from .utils.formatting import formatted_utcnow, pad_timestamp
+from .utils.general_checks import is_int_type
 
 
 class MonitoringValidator:
     def __init__(self):
         pass
 
-    def pre_flight_monitoring_check(self, event, model_info, dataset_info) -> bool:
+    def pre_flight_monitoring_check(
+        self, event, model_info, dataset_info
+    ) -> List[MonitoringViolation]:
 
         violations = []
         # lets track all violations we can find. Its much better to find everything
@@ -168,11 +173,11 @@ class MonitoringValidator:
         elif isinstance(date_str, str):
             try:
                 ts = pad_timestamp(date_str)
-                datetime.strptime(ts, TIMESTAMP_FORMAT)
+                datetime.strptime(ts, constants.TIMESTAMP_FORMAT)
             except Exception:
                 # lets check if its valid ms epoch time
                 is_int, int_val = is_int_type(ts)
-                msg = f'Unable to parse date:{date_str}. Specify date either in millisecs epoch or format:{TIMESTAMP_FORMAT}'
+                msg = f'Unable to parse date:{date_str}. Specify date either in millisecs epoch or format:{constants.TIMESTAMP_FORMAT}'
                 violations.append(
                     MonitoringViolation(MonitoringViolationType.FATAL, msg)
                 )
@@ -331,14 +336,24 @@ class MonitoringValidator:
                 )
 
         if model_info.model_task == ModelTask.BINARY_CLASSIFICATION:
-            if model_info.targets[0].data_type not in [DataType.FLOAT, DataType.INTEGER, DataType.BOOLEAN, DataType.CATEGORY]:
+            if model_info.targets[0].data_type not in [
+                DataType.FLOAT,
+                DataType.INTEGER,
+                DataType.BOOLEAN,
+                DataType.CATEGORY,
+            ]:
                 msg = f'Target columns can only be one of `FLOAT`, `INTEGER`, `BOOLEAN`, or `CATEGORY` for model type:{ModelTask.BINARY_CLASSIFICATION.value}'
                 violations.append(
                     MonitoringViolation(MonitoringViolationType.WARNING, msg)
                 )
             if model_info.targets[0].possible_values is not None:
                 for value in model_info.targets[0].possible_values:
-                    if not isinstance(value, str) and not isinstance(value, float) and not isinstance(value, bool) and not isinstance(value, int):
+                    if (
+                        not isinstance(value, str)
+                        and not isinstance(value, float)
+                        and not isinstance(value, bool)
+                        and not isinstance(value, int)
+                    ):
                         msg = f'Accuracy only works if possible values for target/label columns are strings, floats, integers, or booleans. Current target possible values are {model_info.targets[0].possible_values}'
                         violations.append(
                             MonitoringViolation(MonitoringViolationType.WARNING, msg)
